from typing import Type, Dict, Any, Tuple, Optional
from pydantic import BaseModel, ValidationError, create_model
from pydantic.fields import FieldInfo
from copy import deepcopy
import json

from icecream import ic

def create_partial_model(pydantic_model: Type[BaseModel]) -> Type[BaseModel]:
    '''
    Creates a partial Pydantic model that accepts missing fields.
    This is useful when validating streams of JSON data. 
    Otherwise, Pydantic would raise an error if a field is missing or JSON is malformed.

    Args:
        model (Type[BaseModel]): The Pydantic model.

    Returns:
        Type[BaseModel]: The partial Pydantic model which now accepts missing fields.
    '''
    ic()
    def make_field_optional(field: FieldInfo, default: Any = None) -> Tuple[Any, FieldInfo]:
        ic()
        new = deepcopy(field)
        new.default = default
        new.annotation = Optional[field.annotation] # type: ignore
        return new.annotation, new
    return create_model(
        f'Partial{pydantic_model.__name__}',
        __base__=pydantic_model,
        __module__=pydantic_model.__module__,
        **{
            field_name: make_field_optional(field=field_info)
            for field_name, field_info in pydantic_model.model_fields.items()
        }
    )

def clean_nested_data_with_error_dict(data: Any, pydantic_model: Type[BaseModel]) -> Dict[str, Any]:
    '''
    This function takes the LLM response (data), validates it against the Pydantic model, trys to clean the data and returns the cleaned data.

    This works best with a partial model, as the unvalid data will be removed or set to None.
    To make the Pydantic model partial use the function `partial_model` of the `ValidationManager` class.

    Args:
       data (Any): The data that should be validated or cleaned.
       pydantic_model (Type[BaseModel]): The Pydantic model that should be used for validation.

    Returns:
        Dict[str, Any]: A dictionary with the cleaned data.
    '''
    if isinstance(data, str):
        data = json.loads(data)
    
    try:
        try:
            # Parse the data through the Pydantic model, which validates and constructs a model instance
            model_instance = pydantic_model.model_validate(data)
            ic()
            # Convert the validated model instance back to a dictionary, excluding any undefined model fields
            cleaned_data = model_instance.model_dump(exclude_unset=True)
            ic()
            return cleaned_data
        except ValidationError as e:
            error_dict = e.errors(include_url=False)
            ic()
            # Recursive function to navigate through the nested data structure and set values to None or remove them
            def set_nested_value(data: Any, loc: tuple, error_type: str):
                for key in loc[:-1]:
                    if isinstance(data, dict):
                        data = data[key]
                    elif isinstance(data, list):
                        data = data[int(key)]
                if error_type == 'extra_forbidden':
                    if isinstance(data, dict):
                        del data[loc[-1]]
                    elif isinstance(data, list):
                        del data[int(loc[-1])]
                else:
                    if isinstance(data, dict):
                        data[loc[-1]] = None
                    elif isinstance(data, list):
                        data[int(loc[-1])] = None
            ic()
            # Iterate over each error and handle it
            for item in error_dict:
                loc = item['loc']
                error_type = item['type']
                set_nested_value(data, loc, error_type)
            ic()
            return data
    except Exception as e:
        ic()
        raise e
    


from pydantic import BaseModel, ValidationError
from typing import Dict, Any, Type
import json
from fastapi.encoders import jsonable_encoder
from partial_json_parser import loads
from partial_json_parser.core.options import STR, OBJ
from promptools import extractors
from icecream import ic

from .cleaner import clean_nested_data_with_error_dict, create_partial_model

class ValidationManager:
    '''
    The `ValidationManager` class provides several functions to make sure the responses of LLMs are validated against a given Pydantic model.
    '''            

    ####################
    # HELPER FUNCTION
    ####################
    def add_error_log_to_final_response(
        self,
        response: Dict[str, Any],
        error_message: ValidationError,
        raw_message: Dict[str, Any]
    ) -> Dict[str, Any]:
        '''
        This adds the error message to the final response, with the final validation.

        Args:
            response: the response from the final validation
            error_message: the error message from the first validation
            raw_message: the original message

        Returns:
            the updated response with the error message and the raw message added
        '''
        raw_message = {
            'role': 'assistant',
            'content': json.dumps(raw_message)
        }
        response['raw_message'] = raw_message 
        if error_message is True:
            ic()
            response['validation_error'] = None
            return response
        else:
            ic()
            response['validation_error'] = str(error_message)
            return response

    ####################
    # VALIDATION FUNCTIONS
    ####################
    def validate_for_error_message(
        self,
        response: Dict[str, Any],
        pydantic_model: Type[BaseModel],
    ) -> bool | Type[ValidationError]:
        '''
        This validates the final response from a chat completion (with or without streaming).

        Args:
            response: the final response of the chat completion function
            pydantic_model: the pydantic model for validate the final response

        Returns:
            `True` or `ValidationError`
        '''
        data = json.dumps(response['message']['content'])
        parsed_chunk_dict = json.loads(data)
        ic()
        try:
            if pydantic_model.model_validate(obj=parsed_chunk_dict):
                ic(True)
                return True
        except ValidationError as e:
            ic()
            return e.errors(include_url=False)
    
    def validate_partial_model(
        self,
        response: Dict[str, Any],
        pydantic_model: Type[BaseModel]
    ) -> Dict[str, Any]:
        '''
        This function will take the Pydantic model and parse it into a partial model.
        Then the final response will be compared with the partial model and gets cleaned when still some `ValidationError` occur.
        The response is a cleand version of the final response.

        To use this function the argument `allow_partial` must be set to `True`.

        Args:
            response: the final response of the chat completion function
            pydantic_model: the pydantic model to be used to create the partial model

        Returns:
            Dict[str, Any]: the cleand version of the final response
        '''
        data = response['message']['content']
        parsed_chunk_dict = json.loads(data)
        partial_model = create_partial_model(pydantic_model=pydantic_model)
        ic()
        try:
            #cleaned_data = self.clean_data(data=parsed_chunk_dict, model=partial_model)
            cleaned_data = clean_nested_data_with_error_dict(data=parsed_chunk_dict, pydantic_model=partial_model)
            ic()
            response['message']['content'] = jsonable_encoder(cleaned_data)
            return response
        except ValidationError as e:
            ic()
            raise e
        
    ####################
    # VALIDATION OF CHAT COMPLETION
    ####################
    def validate_chat_completion(
        self,
        response: Dict[str, Any],
        pydantic_model: Type[BaseModel],
    ) -> Dict[str, Any]:
        """
        Validates the response from a chat completion request, ensuring the content is valid according to the provided Pydantic model.
        
        Args:
            response (Dict[str, Any]): The response dictionary from the chat completion request.
            pydantic_model (Type[BaseModel]): The Pydantic model to validate the response against.
        
        Returns:
            Dict[str, Any]: The validated response dictionary, with the content field updated to match the Pydantic model.
        """
        data = response['message']['content']
        parsed_chunk_dict = json.loads(data)
        try:
            valid_data = pydantic_model.model_validate(parsed_chunk_dict)
            ic()
            valid_data.model_dump(exclude_unset=True)
            response['message']['content'] = valid_data
            return response
        except ValidationError as e:
            ic()
            raise e

    ####################
    # VALIDATION OF STREAMS
    ####################
    def validate_chat_completion_with_stream(self, chunk: Dict[str, Any], pydantic_model: Type[BaseModel]) -> Dict[str, Any]:
        ic()
        data = chunk['message']['content']
        fallback_data = {}
        partial_pydantic_model = create_partial_model(pydantic_model=pydantic_model)
        try:
            parsed_chunk = loads(json_string=data, allow_partial=OBJ)
            ic()
            valid_data = extractors.extract_json(text=data, expect=pydantic_model, fallback=partial_pydantic_model(**parsed_chunk).model_dump(), allow_partial=STR | OBJ)
            ic()
            chunk['message']['content'] = jsonable_encoder(valid_data)
        except ValidationError as e:
            ic()
            print('ValidationError', e)
            parsed_chunk = loads(json_string=data, allow_partial=OBJ)
            valid_data = extractors.extract_json(text=data, expect=pydantic_model, fallback=partial_pydantic_model(**fallback_data).model_dump(), allow_partial=OBJ)
            chunk['message']['content'] = jsonable_encoder(valid_data) #custom_encoder=Dict[Any, Any]
            chunk['message']['validation_error'] = str(e.errors(include_url=False))
        except Exception as e:
            ic()
            print('Exception', e)
            parsed_chunk = loads(json_string=data, allow_partial=OBJ)
            chunk['message']['content'] = jsonable_encoder(parsed_chunk)
            chunk['message']['validation_error'] = str(e)
        return chunk



# ollama_instructor_client.py

from icecream import ic
ic.configureOutput(prefix='ollama-instrutor | ')

import rich
import ollama
from typing import Type, Any, Dict, Literal, List, Generator
from pydantic import BaseModel, ValidationError
import json
import re
from fastapi.encoders import jsonable_encoder

from ollama_instructor.prompt_manager import ChatPromptManager
from ollama_instructor.validation_manager import ValidationManager

class OllamaInstructorClient:
    '''
    This class is the client for the ollama instructor. It is a simple wrapper around the ollama client.
    It is the main class that you will use to interact with the ollama instructor.

    Args:
        host (str): The url of the ollama server. Defaults to http://localhost:11434

    Methods:
        `chat_completion`: This method is used for chat completion without streaming.
        `chat_completion_with_streaming`: This method is used for chat completion with streaming.
    
    Both Methods handle your prompts, the LLM's response and its validation identically. For more information look into the docs of both methods.

    A characteristic of ollama instructor is that it makes sure that you get a valid JSON in the structur of the basic Pydantic model you send with the request.
    But, in case of validation failure, it will get rid of the invalid parts of the LLM's response. See the documentation of ollama instructor for more information on this topic.

    Information:
    The ollama instructor will return a modified version of Ollama's response dict. The modifications are as follows:
    - `validation_errors`: This will contain the `ValidationError`raised during Pydantic's validation, if the response of the LLM is invalid.
    - `raw_response`: This will contain the raw response from the LLM.

    
    Example:
    ```Python
    from ollama_instructor.ollama_instructor_client import OllamaInstructorClient
    from pydantic import BaseModel

    class Person(BaseModel):
        name: str
        age: int

    client = OllamaInstructorClient(host="http://localhost:11434") # by default the host is http://localhost:11434
    
    response = client.chat_completion(
        model='llama3',
        pydantic_model=Person,
        messages=[
            {
                'role': 'user',
                'content': 'Jason is 21 years old.'
            }
        ]
    )
    
    print(response)
    ```
    '''
    def __init__(self, host: str = 'http://localhost:11434', debug: bool = False):
        self.ollama_client = ollama.Client(host=host)
        self.validation_manager = ValidationManager()
        self.chat_prompt_manager = ChatPromptManager()
        self.chat_history: List[Dict[str, Any]] = []
        self.retry_counter: int = None
        self.validation_error: bool | str | dict | None = None

        if debug is False:
            ic.disable() 

    ####################
    # CHAT COMPLETION
    ####################
    def chat_completion(self, pydantic_model: Type[BaseModel], messages: List[Dict[str, Any]], model: str, retries: int = 3, format: Literal['', 'json'] = 'json', allow_partial: bool = False, **kwargs):
        '''
        Create a chat completion with the LLM and validate the response with the provided Pydantic model.
        
        **This method is not for streaming.** Use `chat_completion_with_stream` instead.

        Args:
            pydantic_model (Type[BaseModel]): The Pydantic model for validation.
            messages (List[Dict[str, Any]]): The chat messages. ATTENTION: if your messages contain a system prompt, ollama-instructor will not provide its own system prompt. MAKE SURE to instrcut the LLM properly to your needs!
            model (str): The model to use for chatting.
            retries (int): The number of retries if the response is not valid. This will lead into another request with further instructions on the validation error of the previous response.
            format (Literal['', 'json']): The format of the response. By default it is 'json'. It's recommended to use 'json'.
            allow_partial (bool): If True, the response can be partial. Otherwise, the response must contain all fields of the Pydantic model. Have in mind that you can get an uncomplete response when set to `True`.
            **kwargs: Additional arguments to pass to the LLM client of `Ollama`.

        Returns:
            Dict[str, Any]: The Dict contains all key-value-pairs known from `Ollama` and some additional like:
                - the last validation error if there is one
                - the left retries
                - the raw response (the dictionary of 'message') of the LLM before the validation does it's magic

        Example:
            ```Python
            from ollama_instructor.ollama_instructor_client import OllamaInstructorClient
            from pydantic import BaseModel
            from enum import Enum

            client = OllamaInstructorClient(host='http://localhost:11434')

            class Gender(Enum):
                MALE = 'male'
                FEMALE = 'female'

            class Person(BaseModel):
                name: str
                age: int
                gender: Gender

            response = client.chat_completion(
                pydantic_model=Person,
                messages=[
                    {
                        'role': 'user',
                        'content': 'Jason is 45 years old.'
                    }
                ],
                model='mistral:7b-instruct',
            )
            
            print(response['message']['content'])
            # >>> Output example:
            # {'name': 'Jason', 'age': 45, 'gender': 'male'}
            ```
        '''
        self.retry_counter = retries
        system_and_user_prompt = self.chat_prompt_manager.create_chat_prompt_for_json(pydantic_model=pydantic_model, messages=messages)
        ic()
        self.chat_history = system_and_user_prompt
        while self.validation_error is not None or self.retry_counter >= 0: 
            ic(self.retry_counter)
            if self.validation_error is not None:
                ic()
                self.chat_history.append(self.chat_prompt_manager.error_guidance_prompt(validation_error=self.validation_error))
            ic(self.chat_history)
            try:
                ic()
                messages = []
                messages.append(self.chat_history[0])
                messages.append(self.chat_history[1])
                if self.retry_counter != retries:
                    messages.append(self.chat_history[-2])
                    messages.append(self.chat_history[-1])
                ic()
                response = self.ollama_client.chat(
                    model=model,
                    messages=messages,
                    format=format,
                    stream=False,
                    **kwargs
                )
                ic(type(response['message']['content']))
                ic(response['message']['content'])

                content = jsonable_encoder(response['message']['content'])

                # in some cases the llm response json with text at the end. This makes sure the json is extracted from the response.
                response['message']['content'] = json.loads(re.search(pattern=r'\{.*\}', string=content, flags=re.DOTALL).group())
                
                ic(type(response['message']['content']))
                self.validation_error = self.validation_manager.validate_for_error_message(response=response, pydantic_model=pydantic_model)
                ic()
                response = self.validation_manager.add_error_log_to_final_response(response=response, error_message=self.validation_error, raw_message=response['message']['content'])
                ic()
                content_as_str = json.dumps(response['message']['content']) # ?
                chat_history = response # ?
                chat_history['message']['content'] = content_as_str # ?
                self.chat_history.append(response['message'])
                if self.validation_error is not True:
                    if self.retry_counter == 0:
                        ic()
                        if allow_partial:
                            ic(allow_partial)
                            try:
                                response = self.validation_manager.validate_partial_model(response=response, pydantic_model=pydantic_model)
                                ic()
                                response['retries_left'] = self.retry_counter
                                return response
                            except ValidationError as e:
                                raise e
                        else:
                            ic(allow_partial)
                            try:
                                response = self.validation_manager.validate_chat_completion(response=response, pydantic_model=pydantic_model)
                                ic(response)
                                response['retries_left'] = self.retry_counter
                                return response
                            except ValidationError as e:
                                ic()
                                raise e
                    else:
                        ic()
                        self.retry_counter -= 1
                else:
                    ic()
                    response['message']['content'] = json.loads(response['message']['content'])
                    response['retries_left'] = self.retry_counter
                    return response
            except Exception as e:
                ic()
                raise e

    ####################
    # CHAT COMPLETION WITH STREAM
    ####################
    def chat_completion_with_stream(self, pydantic_model: Type[BaseModel], messages: List[Dict[str, Any]], model: str, retries: int = 3, format: Literal['', 'json'] = 'json', allow_partial: bool = False, **kwargs) -> Generator[Dict[str, Any], None, None]: 
        '''
        Chat with the model and validate the response with the provided Pydantic model.

        **This method is for streaming.** Use `chat_completion` instead.

        Args:
            pydantic_model (Type[BaseModel]): The Pydantic model for validation.
            messages (List[Dict[str, Any]]): The chat messages.
            model (str): The LLM model.
            retries (int): The number of retries if the response is not valid.
            format (Literal['', 'json']): The format of the response.
            **kwargs: Additional arguments to pass to the LLM client of `Ollama`.

        Returns:
            Generator[Dict[str, Any], None, None]: The generated response.

        Example:
            ```Python
            from ollama_instructor.ollama_instructor_client import OllamaInstructorClient
            from pydantic import BaseModel
            from enum import Enum

            client = OllamaInstructorClient(host='http://localhost:11434')

            class Gender(Enum):
                MALE = 'male'
                FEMALE = 'female'

            class Person(BaseModel):
                name: str
                age: int
                gender: Gender

            response = client.chat_completion_with_stream(
                pydantic_model=Person,
                messages=[
                    {
                        'role': 'user',
                        'content': 'Jason is 45 years old.'
                    }
                ],
                model='mistral:7b-instruct',
            )
            
            for chunk in response:
                print(chunk['message']['content'])
            # >>> Output example:
            # {'name': 'Jason', 'age': 45, 'gender': 'male'}
        '''
        self.retry_counter = retries
        system_and_user_prompt = self.chat_prompt_manager.create_chat_prompt_for_json(pydantic_model=pydantic_model, messages=messages)
        self.chat_history = system_and_user_prompt
        # start the while loop
        while self.validation_error is not None or self.retry_counter >= 0:
            if self.validation_error is not None:
                self.chat_history.append(self.chat_prompt_manager.error_guidance_prompt(validation_error=self.validation_error))
            expanding_response = ""
            try:
                ic()
                messages = []
                messages.append(self.chat_history[0])
                messages.append(self.chat_history[1])
                if self.retry_counter != retries:
                    messages.append(self.chat_history[-2])
                    messages.append(self.chat_history[-1])
                ic()
                ic(messages)
                ic(type(messages))
                response = self.ollama_client.chat(
                    model=model,
                    messages=messages,
                    format=format,
                    stream=True,
                    **kwargs
                )
                for chunk in response:
                    expanding_response += chunk['message']['content']
                    chunk['message']['content'] = expanding_response
                    # Validation block of chunks
                    chunk = self.validation_manager.validate_chat_completion_with_stream(chunk=chunk, pydantic_model=pydantic_model)
                    self.validation_error = self.validation_manager.validate_for_error_message(response=chunk, pydantic_model=pydantic_model)
                    chunk = self.validation_manager.add_error_log_to_final_response(response=chunk, error_message=self.validation_error, raw_message=expanding_response)
                    yield chunk

                # Append chat_history
                self.chat_history += [chunk['message']]
                # turn the final response into a str
                chunk['message']['content'] = json.dumps(chunk['message']['content'])
                # Validate the final response
                if self.validation_error is not True:
                    if self.retry_counter == 0:
                        ic()
                        if allow_partial:
                            ic(allow_partial)
                            try:
                                chunk = self.validation_manager.validate_partial_model(response=chunk, pydantic_model=pydantic_model)
                                ic()
                                chunk['retries_left'] = self.retry_counter
                                yield chunk
                                return chunk
                            except ValidationError as e:
                                raise e
                        else:
                            ic(allow_partial)
                            try:
                                chunk = self.validation_manager.validate_chat_completion(response=chunk, pydantic_model=pydantic_model)
                                ic(chunk)
                                chunk['retries_left'] = self.retry_counter
                                yield chunk
                                return chunk
                            except ValidationError as e:
                                ic()
                                raise e
                    else:
                        ic()
                        self.retry_counter -= 1
                else:
                    ic()
                    chunk['message']['content'] = json.loads(chunk['message']['content'])
                    chunk['retries_left'] = self.retry_counter
                    yield chunk
                    return chunk
                   
            except Exception as e:
                ic()
                ic(e)
                raise e
            


from pydantic import BaseModel
from typing import List, Dict, Type, Any
import json

class ChatPromptManager:
    def __init__(self):
        self.message_history: List

    ####################
    # BASIC PROMPT
    ####################
    def basic_prompt(self, pydantic_model: Type[BaseModel]):
        '''
        Creates a basic prompt for the LLM based on the given Pydantic model.

        Args:
            pydantic_model (Type[BaseModel]): The Pydantic model.

        Returns:
            str: The basic prompt.
        '''
        default_system_prompt = f'''
You are the world class algorithm for JSON responses. You will get provided a JSON schema of a Pydantic model. Your task is to extract those in this JSON schema specified properties out of a given text or image (or its context) and return a VALID JSON response adhering to the JSON schema.
    \nHere is the JSON schema: {pydantic_model.model_json_schema()}.
    \n\nYou WILL return the instance of the JSON schema with the CORRECT extracted data, NOT the JSON schema itself. The instance of the JSON schema has the following fields to extract the data for: {list(pydantic_model.model_fields.keys())}.
        '''
        return default_system_prompt
    
    ####################
    # ERROR GUIDANCE PROMPT
    ####################
    def error_guidance_prompt(self, validation_error: bool | str | dict) -> Dict[str, Any]:
        '''
        This function creates an error guidance prompt for the LLM.

        Args:
            validation_error (bool | str | dict): The validation error.

        Returns:
            Dict[str, Any]: The error guidance prompt as a message (dict).
        '''
        error_guidance_prompt = {
                                    'role': 'system',
                                    'content': f'The last response raised the following validation error: {json.dumps(validation_error)}. Response with the corrected JSON and fill in the correct data while adhering the context and the JSON schema above!'
                                }
        return error_guidance_prompt
    
    ####################
    # PROMPT HELPER FUNCTION
    ####################
    def create_chat_prompt_for_json(
            self,
            pydantic_model: BaseModel,
            messages: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        '''
        Creates a chat prompt for the LLM based on the given Pydantic model.
        It will contain the system prompt as well, to instruct the LLM to response the instance of the JSON schema.

        ATTENTION: If a system prompt already exists in the messages list, ollama-instructor will not add its OWN basic_prompt.

        Args:
            pydantic_model (Type[BaseModel]): The Pydantic model.
            messages (List[Dict[str, Any]]): The chat messages.

        Returns:
            List[Dict[str, Any]]: The chat messages.
        '''
        chat_messages = messages
        basic_prompt = self.basic_prompt(pydantic_model=pydantic_model)
        if chat_messages[0]['role'] == 'user':
            chat_messages.insert(0, {"role": "system", "content": basic_prompt})
        #elif chat_messages[0]['role'] == 'system':
            #chat_messages.insert(1, {"role": "system", "content": basic_prompt})
        elif chat_messages[0]['role'] == 'system':
            pass
        #else:
            #chat_messages.insert(0, {"role": "system", "content": basic_prompt})
        self.message_history = chat_messages
        return messages